//+------------------------------------------------------------------+
//|  9:51 AM 4/16/2016                            Insertion_Sort.mqh |
//|                               copyright  2016, William H Roeder |
//|                                        mailto:WHRoeder@yahoo.com |
//+------------------------------------------------------------------+
#define INDEX uint
// copy_backward
template <typename Ti, typename To>
static INDEX   copy_backward(const Ti& inp[], INDEX iBeg, INDEX iEnd,
                                   To& out[], INDEX oEnd){
   while(iBeg != iEnd)  out[--oEnd] = inp[--iEnd];
   return oEnd;
}
template <typename T>
static void    unguarded_linear_insert(T& arr[], INDEX iEmpty, T value){
   INDEX    iNext = iEmpty;
   for(;;){
      T  next  = arr[--iNext];
      if(!(value < next))  break;
      arr[iEmpty] = next;
      iEmpty = iNext;
   }
   arr[iEmpty] = value;
}
template <typename T, typename BinaryPredicate>
static void    unguarded_linear_insert(T& arr[], INDEX iEmpty, T value,
                                       BinaryPredicate pred){
   INDEX    iNext = iEmpty;
   for(;;){
      T  next  = arr[--iNext];
      if(!pred.is_before(value, next) )   break;
      arr[iEmpty] = next;
      iEmpty = iNext;
   }
   arr[iEmpty] = value;
}
template <typename T>
static void    linear_insert(T& arr[], INDEX iBeg, INDEX iPlace){
   T     value = arr[iPlace];
   if(value < arr[iBeg]){
      copy_backward(arr, iBeg, iPlace, arr, iPlace + 1);
      arr[iBeg] = value;
   }
   else
      unguarded_linear_insert(arr, iPlace, value);
}
template <typename T, typename BinaryPredicate>
static void    linear_insert(T& arr[], INDEX iBeg, INDEX iPlace
                             BinaryPredicate pred){
   T     value = arr[iPlace];
   if(pred.is_before(value, arr[iBeg]) ){
      copy_backward(arr, iBeg, iPlace, arr, iPlace + 1);
      arr[iBeg] = value;
   }
   else
      unguarded_linear_insert(arr, iPlace, value, pred);
}
template <typename T>
static void    insertion_sort(T& arr[], INDEX iBeg, INDEX iEnd){
   if(iBeg == iEnd)  return;
   for(INDEX iPlace = iBeg + 1; iPlace != iEnd; ++iPlace)
      linear_insert(arr, iBeg, iPlace);
}
template <typename T, typename BinaryPredicate>
static void    insertion_sort(T& arr[], INDEX iBeg, INDEX iEnd,
                              BinaryPredicate pred){
   if(iBeg == iEnd) return;
   for(INDEX iPlace = iBeg + 1; iPlace != iEnd; ++iPlace)
      linear_insert(arr, iBeg, iPlace, pred);
}
